from flask import Flask, request
from flask_babel import Babel
from flask_cors import CORS

# Initialize Flask Application
app = Flask(__name__)

# Include extra route files
from . import flask_app  # noqa: F401,E402

# Handle CORS call
CORS(app, resources={r"/*": {"origins": "*"}})

# Handle translation via babel
app.config["BABEL_DEFAULT_LOCALE"] = "en"
babel = Babel(app)


def get_locale():
    # Use the browser's language preferences to select an available translation
    available_translations = [
        str(translation) for translation in babel.list_translations()
    ]
    for lang in request.accept_languages.values():
        if lang[:2] in available_translations:
            return lang[:2]


babel.init_app(app, locale_selector=get_locale)



from .models.device_abstract import DeviceAbstract
from .scale_dialog06 import state_machine


class DeviceScale(DeviceAbstract):
    device_type = "scale"

    def __init__(self, interface, options):
        DeviceAbstract.__init__(self, interface, options)
        self.last_weight = 0.0

    def read_weight(self, unit_price, tare):
        with self._get_serial() as serial:
            sm = state_machine.Dialog06Machine(
                serial,
                self.get_option("polynomial", int, required=True),
                unit_price,
                tare,
            )

            while not sm.current_state.final:
                sm.continue_communication()

        # In a dialog06 context, if the scale return 21 (no change since the last weight)
        # The scale doesn't provide the weight.
        # in that 'valid use case', we return the last weight
        if sm.last_scale_answer.error_number != "21":
            self.last_weight = sm.last_scale_answer.weight

        return {
            "weight": self.last_weight,
            "error_number": sm.last_scale_answer.error_number,
            "error_description": sm.last_scale_answer.error_description,
        }



import uuid
from datetime import datetime


class DeviceTask:
    def __init__(self, data):
        self.create_date = datetime.now()
        self.uuid = uuid.uuid4()
        self.data = data
        self.error = False



import simplejson
from flask import jsonify, request
from loguru import logger

from ..app import app
from ..interface import interface


@app.route("/hw_proxy/payment_terminal_transaction_start", methods=["POST", "PUT"])
@logger.catch
def payment_terminal_transaction_start():
    logger.trace("payment_terminal_transaction_start()")
    data = request.json.get("params", {}).get("payment_info", {})
    interface.log_data(data, "payment", "DEBUG")

    if type(data) is str:
        data = simplejson.loads(data)

    amount = float(data.get("amount"))
    currency_iso = data.get("currency_iso")

    result = interface.device_payment_push_amount(amount, currency_iso)
    return jsonify(jsonrpc="2.0", result=result)



from statemachine import State, StateMachine

from .scale_command import send_ask_status, send_ask_weight, send_checksums, send_info


class Dialog06Machine(StateMachine):
    def __init__(self, device, polynomial, unit_price, tare=0.0, extra_text=""):
        # Device configuration
        self._device = device
        self.polynomial = polynomial

        # Data to send to the scale
        self.unit_price = unit_price
        self.tare = tare
        self.extra_text = extra_text

        # contain the last analyzed answer from the scale
        self.last_scale_answer = False

        return super().__init__()

    # #############################
    # Public Section
    # #############################

    def continue_communication(self):
        event_name = f"event_{self.last_scale_answer.answer_type}"
        self.send(event_name)

    # #############################
    # State Definition
    # #############################

    state_send_info = State(initial=True)

    state_send_checksums = State()

    state_ask_weight = State()

    state_ask_error = State()

    state_weight_recovered = State(final=True)

    state_error = State(final=True)

    state_unknown_error = State(final=True)

    # #############################
    # Event Definition
    # #############################

    event_record_ACK = state_send_info.to(state_ask_weight) | state_send_checksums.to(
        state_send_info
    )

    event_record_NAK = state_send_info.to(state_ask_error) | state_ask_weight.to(
        state_ask_error
    )

    event_record_11_checksums_valid = state_ask_weight.to(state_ask_weight)

    event_record_02_weight = state_ask_weight.to(state_weight_recovered)

    event_record_09_error = state_ask_error.to(state_error)

    event_record_09_no_error = state_ask_error.to(state_send_info)

    event_record_11_checksums_request = state_send_info.to(state_send_checksums)

    event_unknown_error = state_send_info.to(state_unknown_error)

    # #############################
    # State Section
    # #############################

    def on_enter_state_send_info(self):
        self.last_scale_answer = send_info(
            self._device, self.unit_price, self.tare, self.extra_text
        )

    def on_enter_state_send_checksums(self):
        self.last_scale_answer = send_checksums(
            self._device,
            self.polynomial,
            self.last_scale_answer.checksums_d0,
            self.last_scale_answer.checksums_z,
        )

    def on_enter_state_ask_weight(self):
        self.last_scale_answer = send_ask_weight(self._device)

    def on_enter_state_ask_error(self):
        self.last_scale_answer = send_ask_status(self._device)



import time
from queue import Queue
from threading import Thread

from .device_abstract import DeviceAbstract
from .device_task import DeviceTask


class DeviceAbstractWithQueue(DeviceAbstract, Thread):
    has_queue = True

    def __init__(self, interface, options, delay=0, max_queue_size=0):
        """Initialize a new Device.
        delay: time between processing two messages, in seconds.
        """
        DeviceAbstract.__init__(self, interface, options)
        Thread.__init__(self)
        self.errored_tasks = []
        self._queue = Queue(max_queue_size)
        self.delay = delay

    @property
    def queue_size(self):
        return self._queue.qsize()

    @property
    def max_queue_size(self):
        return self._queue.maxsize

    @property
    def errored_tasks_qty(self):
        return len(self.errored_tasks)

    # ###########################
    # Thread / Task Section
    # ###########################
    def run(self):
        while True:
            if self._usb_device:
                current_size = self._queue.qsize()
                if current_size:
                    self._logger("INFO", f"Process one of the {current_size} task(s).")
                else:
                    self._logger("INFO", "Waiting for message ...")
                task = self._queue.get(True)
                self._process_task_abstract(task)
                time.sleep(self.delay)
            else:
                time.sleep(1)

    def _process_task_abstract(self, task):
        self._logger("DEBUG", f"Processing task {task.uuid} ...")
        try:
            if self.use_serial:
                with self._get_serial() as serial:
                    self._process_task_serial(serial, task)
            else:
                self._process_task(task)
            if task in self.errored_tasks:
                self.errored_tasks.remove(task)
            self._logger("SUCCESS", f"Task processed. UUID: {task.uuid}.")

        except Exception as e:
            self._logger("ERROR", f"Task {task.uuid}. Error{e}")
            task.error = e
            if task not in self.errored_tasks:
                self._logger("INFO", f"Reenqueing task. UUID: {task.uuid} ...")
                self.errored_tasks.append(task)
                self._queue.put(task)
            if self._usb_device:
                self._interface._hook_usb_device_removed(self._usb_device, e)

    def add_task(self, data):
        if self.max_queue_size:
            while self._queue.qsize() >= self._queue.maxsize:
                self._logger("WARNING", "Removing obsolete task in queue.")
                self._queue.get_nowait()
        task = DeviceTask(data)
        self._queue.put(task)
        self._logger(
            "INFO",
            f"Task Enqueued ({self._queue.qsize()}/{self._queue.maxsize})."
            f" UUID: {task.uuid}",
        )



import logging
import os
import sys
from pathlib import Path

import click
import importlib_resources
import simplejson
from click_loglevel import LogLevel
from gevent import pywsgi
from geventwebsocket.handler import WebSocketHandler
from loguru import logger

import odoo_driver

from .app import app
from .interface import interface

# Define loguru specific level for logging
# to make click_loglevel working
logging.addLevelName(5, "TRACE")
logging.addLevelName(25, "SUCCESS")


@click.command()
@click.option(
    "-a",
    "--address",
    type=click.STRING,
    default="0.0.0.0",
    show_default=True,
    help="Address on which the web service will be exposed",
)
@click.option(
    "-p",
    "--port",
    type=click.INT,
    default=8069,
    show_default=True,
    help="Port on which the web service will be exposed",
)
@click.option(
    "--secure/--unsecure",
    default=True,
    show_default=True,
    help="Option 'secure' exposes web service on https."
    " Option 'unsecure' exposes web service on http.",
)
@click.option(
    "-r",
    "--refresh-devices-delay",
    type=click.INT,
    default=1,
    show_default=True,
    help="Interval in seconds between two device refreshes",
)
@click.option(
    "--options",
    type=click.STRING,
    default="{}",
    show_default=True,
    help="""Extra JSON options to pass to devices. Exemple:"
    " --options '{"scale": {"polynomial": "123456"}}'
    """,
)
@click.option(
    "-l",
    "--log-level",
    type=LogLevel(extra=["TRACE", "SUCCESS"]),
    default="INFO",
    show_default=True,
)
@click.option(
    "-f",
    "--log-folder",
    type=click.Path(
        exists=True, file_okay=False, writable=True, resolve_path=True, path_type=Path
    ),
    show_default=True,
)
@click.version_option(version=odoo_driver.__version__)
def main(log_level, address, port, refresh_devices_delay, secure, log_folder, options):
    logger.success("Lauching odoo-driver ...")

    logger.info(f"Application Path: {os.path.abspath(__file__)}")

    # Log Configuration
    # remove first default sink that is defined to 'INFO'
    logger.remove()

    # Add stdout logger
    logger.add(sys.stderr, level=log_level)

    # Add file logger
    if log_folder:
        log_path = log_folder / "odoo-driver_{time}.log"
        logger.info(f"Log Path: {log_path}")
        logger.add(log_path, level=log_level, rotation="1 week", enqueue=True)

    # Initialize Interface
    interface.initialize(
        refresh_devices_delay=refresh_devices_delay,
        options=simplejson.loads(options),
        log_folder=log_folder,
    )

    kwargs = {"handler_class": WebSocketHandler}

    # Handle HTTPS, if required
    if secure:
        cert_folder = importlib_resources.files("odoo_driver") / "default_cert"
        server_key_path = cert_folder / "server.key"
        server_cert_path = cert_folder / "server.crt"
        kwargs.update({"keyfile": server_key_path, "certfile": server_cert_path})
    http_server = pywsgi.WSGIServer((address, port), app, **kwargs)
    prefix = secure and "https" or "http"
    logger.success(
        f"Serving odoo-driver (version: {odoo_driver.__version__})"
        f" on {prefix}://{address}:{port} ..."
    )
    http_server.serve_forever()
